@@ -1,18 +1,19 @@
-from typing import List, Callable, Optional, Dict
+from typing import List, Callable, Optional, Dict, Union
+import numpy as np
 import aplr_cpp
+import itertools
 
-FloatVector = List[float]
-FloatMatrix = List[List[float]]
-IntVector = List[int]
-IntMatrix = List[List[int]]
-StrVector = List[str]
+FloatVector = np.ndarray
+FloatMatrix = np.ndarray
+IntVector = np.ndarray
+IntMatrix = np.ndarray
 
 
 class APLRRegressor:
     def __init__(
         self,
         m: int = 3000,
-        v: float = 0.1,
+        v: float = 0.5,
         random_state: int = 0,
         loss_function: str = "mse",
         link_function: str = "identity",
@@ -21,9 +22,9 @@ class APLRRegressor:
         bins: int = 300,
         max_interaction_level: int = 1,
         max_interactions: int = 100000,
-        min_observations_in_split: int = 20,
-        ineligible_boosting_steps_added: int = 10,
-        max_eligible_terms: int = 5,
+        min_observations_in_split: int = 4,
+        ineligible_boosting_steps_added: int = 15,
+        max_eligible_terms: int = 7,
         verbosity: int = 0,
         dispersion_parameter: float = 1.5,
         validation_tuning_metric: str = "default",
@@ -68,11 +69,12 @@ class APLRRegressor:
         monotonic_constraints_ignore_interactions: bool = False,
         group_mse_by_prediction_bins: int = 10,
         group_mse_cycle_min_obs_in_bin: int = 30,
-        early_stopping_rounds: int = 500,
+        early_stopping_rounds: int = 200,
         num_first_steps_with_linear_effects_only: int = 0,
         penalty_for_non_linearity: float = 0.0,
         penalty_for_interactions: float = 0.0,
         max_terms: int = 0,
+        ridge_penalty: float = 0.0001,
     ):
         self.m = m
         self.v = v
@@ -119,6 +121,7 @@ class APLRRegressor:
         self.penalty_for_non_linearity = penalty_for_non_linearity
         self.penalty_for_interactions = penalty_for_interactions
         self.max_terms = max_terms
+        self.ridge_penalty = ridge_penalty
 
         # Creating aplr_cpp and setting parameters
         self.APLRRegressor = aplr_cpp.APLRRegressor()
@@ -179,22 +182,24 @@ class APLRRegressor:
         self.APLRRegressor.penalty_for_non_linearity = self.penalty_for_non_linearity
         self.APLRRegressor.penalty_for_interactions = self.penalty_for_interactions
         self.APLRRegressor.max_terms = self.max_terms
+        self.APLRRegressor.ridge_penalty = self.ridge_penalty
 
     def fit(
         self,
         X: FloatMatrix,
         y: FloatVector,
-        sample_weight: FloatVector = [],
-        X_names: StrVector = [],
-        cv_observations: IntMatrix = [],
-        prioritized_predictors_indexes: IntVector = [],
-        monotonic_constraints: IntVector = [],
-        group: FloatVector = [],
+        sample_weight: FloatVector = np.empty(0),
+        X_names: List[str] = [],
+        cv_observations: IntMatrix = np.empty([0, 0]),
+        prioritized_predictors_indexes: List[int] = [],
+        monotonic_constraints: List[int] = [],
+        group: FloatVector = np.empty(0),
         interaction_constraints: List[List[int]] = [],
-        other_data: FloatMatrix = [],
-        predictor_learning_rates: FloatVector = [],
-        predictor_penalties_for_non_linearity: FloatVector = [],
-        predictor_penalties_for_interactions: FloatVector = [],
+        other_data: FloatMatrix = np.empty([0, 0]),
+        predictor_learning_rates: List[float] = [],
+        predictor_penalties_for_non_linearity: List[float] = [],
+        predictor_penalties_for_interactions: List[float] = [],
+        predictor_min_observations_in_split: List[int] = [],
     ):
         self.__set_params_cpp()
         self.APLRRegressor.fit(
@@ -211,6 +216,7 @@ class APLRRegressor:
             predictor_learning_rates,
             predictor_penalties_for_non_linearity,
             predictor_penalties_for_interactions,
+            predictor_min_observations_in_split,
         )
 
     def predict(
@@ -222,16 +228,16 @@ class APLRRegressor:
             )
         return self.APLRRegressor.predict(X, cap_predictions_to_minmax_in_training)
 
-    def set_term_names(self, X_names: StrVector):
+    def set_term_names(self, X_names: List[str]):
         self.APLRRegressor.set_term_names(X_names)
 
     def calculate_feature_importance(
-        self, X: FloatMatrix, sample_weight: FloatVector = []
+        self, X: FloatMatrix, sample_weight: FloatVector = np.empty(0)
     ) -> FloatVector:
         return self.APLRRegressor.calculate_feature_importance(X, sample_weight)
 
     def calculate_term_importance(
-        self, X: FloatMatrix, sample_weight: FloatVector = []
+        self, X: FloatMatrix, sample_weight: FloatVector = np.empty(0)
     ) -> FloatVector:
         return self.APLRRegressor.calculate_term_importance(X, sample_weight)
 
@@ -242,7 +248,7 @@ class APLRRegressor:
         return self.APLRRegressor.calculate_local_term_contribution(X)
 
     def calculate_local_contribution_from_selected_terms(
-        self, X: FloatMatrix, predictor_indexes: IntVector
+        self, X: FloatMatrix, predictor_indexes: List[int]
     ) -> FloatVector:
         return self.APLRRegressor.calculate_local_contribution_from_selected_terms(
             X, predictor_indexes
@@ -251,13 +257,13 @@ class APLRRegressor:
     def calculate_terms(self, X: FloatMatrix) -> FloatMatrix:
         return self.APLRRegressor.calculate_terms(X)
 
-    def get_term_names(self) -> StrVector:
+    def get_term_names(self) -> List[str]:
         return self.APLRRegressor.get_term_names()
 
-    def get_term_affiliations(self) -> StrVector:
+    def get_term_affiliations(self) -> List[str]:
         return self.APLRRegressor.get_term_affiliations()
 
-    def get_unique_term_affiliations(self) -> StrVector:
+    def get_unique_term_affiliations(self) -> List[str]:
         return self.APLRRegressor.get_unique_term_affiliations()
 
     def get_base_predictors_in_each_unique_term_affiliation(self) -> List[List[int]]:
@@ -294,15 +300,27 @@ class APLRRegressor:
         return self.APLRRegressor.get_main_effect_shape(predictor_index)
 
     def get_unique_term_affiliation_shape(
-        self, unique_term_affiliation: str, max_rows_before_sampling: int = 100000
+        self,
+        unique_term_affiliation: str,
+        max_rows_before_sampling: int = 500000,
+        additional_points: int = 250,
     ) -> FloatMatrix:
         return self.APLRRegressor.get_unique_term_affiliation_shape(
-            unique_term_affiliation, max_rows_before_sampling
+            unique_term_affiliation, max_rows_before_sampling, additional_points
         )
 
     def get_cv_error(self) -> float:
         return self.APLRRegressor.get_cv_error()
 
+    def set_intercept(self, value: float):
+        self.APLRRegressor.set_intercept(value)
+
+    def remove_provided_custom_functions(self):
+        self.APLRRegressor.remove_provided_custom_functions()
+        self.calculate_custom_validation_error_function = None
+        self.calculate_custom_loss_function = None
+        self.calculate_custom_negative_gradient_function = None
+
     # For sklearn
     def get_params(self, deep=True):
         return {
@@ -337,6 +355,7 @@ class APLRRegressor:
             "penalty_for_non_linearity": self.penalty_for_non_linearity,
             "penalty_for_interactions": self.penalty_for_interactions,
             "max_terms": self.max_terms,
+            "ridge_penalty": self.ridge_penalty,
         }
 
     # For sklearn
@@ -351,7 +370,7 @@ class APLRClassifier:
     def __init__(
         self,
         m: int = 3000,
-        v: float = 0.1,
+        v: float = 0.5,
         random_state: int = 0,
         n_jobs: int = 0,
         cv_folds: int = 5,
@@ -359,16 +378,17 @@ class APLRClassifier:
         verbosity: int = 0,
         max_interaction_level: int = 1,
         max_interactions: int = 100000,
-        min_observations_in_split: int = 20,
-        ineligible_boosting_steps_added: int = 10,
-        max_eligible_terms: int = 5,
+        min_observations_in_split: int = 4,
+        ineligible_boosting_steps_added: int = 15,
+        max_eligible_terms: int = 7,
         boosting_steps_before_interactions_are_allowed: int = 0,
         monotonic_constraints_ignore_interactions: bool = False,
-        early_stopping_rounds: int = 500,
+        early_stopping_rounds: int = 200,
         num_first_steps_with_linear_effects_only: int = 0,
         penalty_for_non_linearity: float = 0.0,
         penalty_for_interactions: float = 0.0,
         max_terms: int = 0,
+        ridge_penalty: float = 0.0001,
     ):
         self.m = m
         self.v = v
@@ -395,6 +415,7 @@ class APLRClassifier:
         self.penalty_for_non_linearity = penalty_for_non_linearity
         self.penalty_for_interactions = penalty_for_interactions
         self.max_terms = max_terms
+        self.ridge_penalty = ridge_penalty
 
         # Creating aplr_cpp and setting parameters
         self.APLRClassifier = aplr_cpp.APLRClassifier()
@@ -429,20 +450,22 @@ class APLRClassifier:
         self.APLRClassifier.penalty_for_non_linearity = self.penalty_for_non_linearity
         self.APLRClassifier.penalty_for_interactions = self.penalty_for_interactions
         self.APLRClassifier.max_terms = self.max_terms
+        self.APLRClassifier.ridge_penalty = self.ridge_penalty
 
     def fit(
         self,
         X: FloatMatrix,
-        y: StrVector,
-        sample_weight: FloatVector = [],
-        X_names: StrVector = [],
-        cv_observations: IntMatrix = [],
-        prioritized_predictors_indexes: IntVector = [],
-        monotonic_constraints: IntVector = [],
+        y: List[str],
+        sample_weight: FloatVector = np.empty(0),
+        X_names: List[str] = [],
+        cv_observations: IntMatrix = np.empty([0, 0]),
+        prioritized_predictors_indexes: List[int] = [],
+        monotonic_constraints: List[int] = [],
         interaction_constraints: List[List[int]] = [],
-        predictor_learning_rates: FloatVector = [],
-        predictor_penalties_for_non_linearity: FloatVector = [],
-        predictor_penalties_for_interactions: FloatVector = [],
+        predictor_learning_rates: List[float] = [],
+        predictor_penalties_for_non_linearity: List[float] = [],
+        predictor_penalties_for_interactions: List[float] = [],
+        predictor_min_observations_in_split: List[int] = [],
     ):
         self.__set_params_cpp()
         self.APLRClassifier.fit(
@@ -457,7 +480,10 @@ class APLRClassifier:
             predictor_learning_rates,
             predictor_penalties_for_non_linearity,
             predictor_penalties_for_interactions,
+            predictor_min_observations_in_split,
         )
+        # For sklearn
+        self.classes_ = np.arange(len(self.APLRClassifier.get_categories()))
 
     def predict_class_probabilities(
         self, X: FloatMatrix, cap_predictions_to_minmax_in_training: bool = False
@@ -468,13 +494,13 @@ class APLRClassifier:
 
     def predict(
         self, X: FloatMatrix, cap_predictions_to_minmax_in_training: bool = False
-    ) -> StrVector:
+    ) -> List[str]:
         return self.APLRClassifier.predict(X, cap_predictions_to_minmax_in_training)
 
     def calculate_local_feature_contribution(self, X: FloatMatrix) -> FloatMatrix:
         return self.APLRClassifier.calculate_local_feature_contribution(X)
 
-    def get_categories(self) -> StrVector:
+    def get_categories(self) -> List[str]:
         return self.APLRClassifier.get_categories()
 
     def get_logit_model(self, category: str) -> APLRRegressor:
@@ -489,7 +515,7 @@ class APLRClassifier:
     def get_feature_importance(self) -> FloatVector:
         return self.APLRClassifier.get_feature_importance()
 
-    def get_unique_term_affiliations(self) -> StrVector:
+    def get_unique_term_affiliations(self) -> List[str]:
         return self.APLRClassifier.get_unique_term_affiliations()
 
     def get_base_predictors_in_each_unique_term_affiliation(self) -> List[List[int]]:
@@ -517,6 +543,7 @@ class APLRClassifier:
             "penalty_for_non_linearity": self.penalty_for_non_linearity,
             "penalty_for_interactions": self.penalty_for_interactions,
             "max_terms": self.max_terms,
+            "ridge_penalty": self.ridge_penalty,
         }
 
     # For sklearn
@@ -525,3 +552,66 @@ class APLRClassifier:
             setattr(self, parameter, value)
         self.__set_params_cpp()
         return self
+
+    # For sklearn
+    def predict_proba(self, X: FloatMatrix) -> FloatMatrix:
+        return self.predict_class_probabilities(X)
+
+
+class APLRTuner:
+    def __init__(
+        self,
+        parameters: Union[Dict[str, List[float]], List[Dict[str, List[float]]]] = {
+            "max_interaction_level": [0, 1],
+            "min_observations_in_split": [4, 10, 20, 100, 500, 1000],
+        },
+        is_regressor: bool = True,
+    ):
+        self.parameters = parameters
+        self.is_regressor = is_regressor
+        self.parameter_grid = self._create_parameter_grid()
+
+    def _create_parameter_grid(self) -> List[Dict[str, float]]:
+        items = sorted(self.parameters.items())
+        keys, values = zip(*items)
+        combinations = list(itertools.product(*values))
+        grid = [dict(zip(keys, combination)) for combination in combinations]
+        return grid
+
+    def fit(self, X: FloatMatrix, y: FloatVector, **kwargs):
+        self.cv_results: List[Dict[str, float]] = []
+        best_validation_result = np.inf
+        for params in self.parameter_grid:
+            if self.is_regressor:
+                model = APLRRegressor(**params)
+            else:
+                model = APLRClassifier(**params)
+            model.fit(X, y, **kwargs)
+            cv_error_for_this_model = model.get_cv_error()
+            cv_results_for_this_model = model.get_params()
+            cv_results_for_this_model["cv_error"] = cv_error_for_this_model
+            self.cv_results.append(cv_results_for_this_model)
+            if cv_error_for_this_model < best_validation_result:
+                best_validation_result = cv_error_for_this_model
+                self.best_model = model
+        self.cv_results = sorted(self.cv_results, key=lambda x: x["cv_error"])
+
+    def predict(self, X: FloatMatrix, **kwargs) -> Union[FloatVector, List[str]]:
+        return self.best_model.predict(X, **kwargs)
+
+    def predict_class_probabilities(self, X: FloatMatrix, **kwargs) -> FloatMatrix:
+        if self.is_regressor == False:
+            return self.best_model.predict_class_probabilities(X, **kwargs)
+        else:
+            raise TypeError(
+                "predict_class_probabilities is only possible when is_regressor is False"
+            )
+
+    def predict_proba(self, X: FloatMatrix, **kwargs) -> FloatMatrix:
+        return self.predict_class_probabilities(X, **kwargs)
+
+    def get_best_estimator(self) -> Union[APLRClassifier, APLRRegressor]:
+        return self.best_model
+
+    def get_cv_results(self) -> List[Dict[str, float]]:
+        return self.cv_results
